
# Copyright (C) 2017 GWU Flight Dynamics and Control Laboratory 
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Attitude Kinematics - transformations between attitude representations

This module provides a variety of functions to perform attitude kinematics.
It transforms between a variety of attitude representations, provides functions
for working with attitude, 

Example
-------
Provide some examples of how to use this module

References
----------
[1] M. D. Shuster, A survey of attitude representations, Journal of the Astronautical Sciences, vol. 41, no. 8, pp. 439-517, 1993.
[2] P. C. Hughes, Spacecraft Attitude Dynamics. Dover Publications, 2004.
[3] J. R. Wertz, Spacecraft Attitude Determination and Control, vol. 73. Springer, 1978.

Requirements
------------
List all the required components for this library to work properly.

.. _NumPy Documentation HOWTO:
   https://github.com/numpy/numpy/blob/master/doc/HOWTO_DOCUMENT.rst.txt

"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
from math import floor, ceil

def rot1(angle, form='c'):
    """Euler rotation about first axis

    This computes the rotation matrix associated with a rotation about the first
    axis. It will output matrices assuming column or row format vectors. 

    For example, to transform a vector from reference frame b to reference frame a:

    Column Vectors : a = rot1(angle, 'c').dot(b)
    Row Vectors    : a = b.dot(rot1(angle, 'r'))

    It should be clear that rot1(angle, 'c') = rot1(angle, 'r').T

    Parameters
    ----------
    angle : float
        Angle of rotation about first axis. In radians
    form : str
        Flag to choose between row or column vector convention.

    Returns
    -------
    mat : numpy.ndarray of shape (3,3)
        Rotation matrix
    """
    cos_a = np.cos(angle)
    sin_a = np.sin(angle)
    rot_mat = np.identity(3)

    if form=='c':
        rot_mat[1, 1] = cos_a
        rot_mat[1, 2] = -sin_a
        rot_mat[2, 1] = sin_a
        rot_mat[2, 2] = cos_a
    elif form=='r':
        rot_mat[1, 1] = cos_a
        rot_mat[1, 2] = sin_a
        rot_mat[2, 1] = -sin_a
        rot_mat[2, 2] = cos_a
    else:
        print("Unknown input. 'r' or 'c' for row/column notation.")
        return 1

    return rot_mat

def rot2(angle, form='c'):
    """Euler rotation about second axis

    This computes the rotation matrix associated with a rotation about the first
    axis. It will output matrices assuming column or row format vectors. 

    For example, to transform a vector from reference frame b to reference frame a:

    Column Vectors : a = rot2(angle, 'c').dot(b)
    Row Vectors    : a = b.dot(rot1(angle, 'r'))

    It should be clear that rot2(angle, 'c') = rot2(angle, 'r').T

    Parameters
    ----------
    angle : float
        Angle of rotation about second axis. In radians
    form : str
        Flag to choose between row or column vector convention.

    Returns
    -------
    mat : numpy.ndarray of shape (3,3)
        Rotation matrix
    """
    cos_a = np.cos(angle)
    sin_a = np.sin(angle)
    rot_mat = np.identity(3)

    if form == 'c':
        rot_mat[0, 0] = cos_a
        rot_mat[0, 2] = sin_a
        rot_mat[2, 0] = -sin_a
        rot_mat[2, 2] = cos_a
    elif form == 'r': 
        rot_mat[0, 0] = cos_a
        rot_mat[0, 2] = -sin_a
        rot_mat[2, 0] = sin_a
        rot_mat[2, 2] = cos_a
    else:
        print("Unknown input. 'r' or 'c' for row/column notation.")
        return 1

    return rot_mat 

def rot3(angle, form='c'):
    """Euler rotation about thrid axis

    This computes the rotation matrix associated with a rotation about the third
    axis. It will output matrices assuming column or row format vectors. 

    For example, to transform a vector from reference frame b to reference frame a:

    Column Vectors : a = rot3(angle, 'c').dot(b)
    Row Vectors    : a = b.dot(rot3(angle, 'r'))

    It should be clear that rot3(angle, 'c') = rot3(angle, 'r').T

    Parameters
    ----------
    angle : float
        Angle of rotation about first axis. In radians
    form : str
        Flag to choose between row or column vector convention.

    Returns
    -------
    mat : numpy.ndarray of shape (3,3)
        Rotation matrix
    """
    cos_a = np.cos(angle)
    sin_a = np.sin(angle)
    rot_mat = np.identity(3)

    if form == 'c':
        rot_mat[0, 0] = cos_a
        rot_mat[0, 1] = -sin_a
        rot_mat[1, 0] = sin_a
        rot_mat[1, 1] = cos_a
    elif form == 'r':
        rot_mat[0, 0] = cos_a
        rot_mat[0, 1] = sin_a
        rot_mat[1, 0] = -sin_a
        rot_mat[1, 1] = cos_a
    else:
        print("Unknown input. 'r' or 'c' for row/column notation.")
        return 1

    return rot_mat
 
def dcm2body313(dcm):
    """Convert DCM to body Euler 3-1-3 angles

    """
    theta = np.zeros(3)

    theta[0] = np.arctan2(dcm[2, 0], dcm[2, 1])
    theta[1] = np.arccos(dcm[2, 2])
    theta[2] = np.arctan2(dcm[0, 2], -dcm[1, 2])

    return theta

def body313todcm(theta):
    dcm = np.zeros((3,3))
    st1 = np.sin(theta[0])
    st2 = np.sin(theta[1])
    st3 = np.sin(theta[2])
    ct1 = np.cos(theta[0])
    ct2 = np.cos(theta[1])
    ct3 = np.cos(theta[2])

    dcm[0,0] = ct1*ct3-ct2*st1*st3
    dcm[0,1] = -ct3*st1-ct1*ct2*st3
    dcm[0,2] = st3*st2
    dcm[1,0] = ct1*st3+ct3*ct2*st1
    dcm[1,1] = ct1*ct3*ct2-st1*st3
    dcm[1,2] = -ct3*st2
    dcm[2,0] = st1*st2
    dcm[2,1] = ct1*st2
    dcm[2,2] = ct2

    return dcm

def body313dot(theta,Omega):
    st1 = np.sin(theta[0])
    st2 = np.sin(theta[1])
    st3 = np.sin(theta[2])
    ct1 = np.cos(theta[0])
    ct2 = np.cos(theta[1])
    ct3 = np.cos(theta[2])

    mat = np.array([[-ct2*st1,-ct1*ct2, st2],
                    [st2*ct1, -st2*st1, 0],
                    [st1,ct1,0]])

    theta_dot = 1/st2*mat.dot(Omega)

    return theta_dot

def body313dot_to_ang_vel(theta,theta_dot):
    st1 = np.sin(theta[0])
    st2 = np.sin(theta[1])
    st3 = np.sin(theta[2])
    ct1 = np.cos(theta[0])
    ct2 = np.cos(theta[1])
    ct3 = np.cos(theta[2])

    mat = np.array([[0,ct1,st1*st2],
                    [0,-st1, ct1*st2],
                    [1,0,ct2]])

    Omega = mat.dot(theta_dot)
    return Omega

def hat_map(vec):
    """Return that hat map of a vector
    
    Inputs: 
        vec - 3 element vector

    Outputs:
        skew - 3,3 skew symmetric matrix

    """
    vec = np.squeeze(vec)
    skew = np.array([
                    [0, -vec[2], vec[1]],
                    [vec[2], 0, -vec[0]],
                    [-vec[1], vec[0], 0]])

    return skew

def vee_map(skew):
    """Return the vee map of a vector

    """

    vec = 1/2 * np.array([skew[2,1] - skew[1,2],
                          skew[0,2] - skew[2,0],
                          skew[1,0] - skew[0,1]])

    return vec

def dcmtoquat(dcm):
    """Convert DCM to quaternion
    
    This function will convert a rotation matrix, also called a direction
    cosine matrix into the equivalent quaternion.

    Parameters:
    ----------
    dcm - (3,3) numpy array
        Numpy rotation matrix which defines a rotation from the b to a frame

    Returns:
    --------
    quat - (4,) numpy array
        Array defining a quaterion where the quaternion is defined in terms of
        a vector and a scalar part.  The vector is related to the eigen axis
        and equivalent in both reference frames [x y z w]
    """
    quat = np.zeros(4)
    quat[-1] = 1/2*np.sqrt(np.trace(dcm)+1)
    quat[0:3] = 1/4/quat[-1]*vee_map(dcm-dcm.T)

    return quat

def quattodcm(quat):
    """Convert quaternion to DCM

    This function will convert a quaternion to the equivalent rotation matrix,
    or direction cosine matrix.

    Parameters:
    ----------
    quat - (4,) numpy array
        Array defining a quaterion where the quaternion is defined in terms of
        a vector and a scalar part.  The vector is related to the eigen axis
        and equivalent in both reference frames [x y z w]

    Returns:
    --------
    dcm - (3,3) numpy array
        Numpy rotation matrix which defines a rotation from the b to a frame
    """

    dcm = (quat[-1]**2-np.inner(quat[0:3], quat[0:3]))*np.eye(3,3) + 2*np.outer(quat[0:3],quat[0:3]) + 2*quat[-1]*hat_map(quat[0:3])

    return dcm

def quatdot(quat, Omega):

    quat_dot = np.zeros(4)
    quat_dot[0:3] = 1/2*(quat[-1]*np.eye(3,3) + hat_map(quat[0:3])).dot(Omega)
    quat_dot[-1] = -1/2*Omega.dot(quat[0:3])

    return quat_dot

def quatdot_ang_vel(quat, quat_dot):
    q = quat[0:3]
    q4 = quat[-1]
    qd = quat_dot[0:3]
    q4d = quat_dot[-1]

    Omega = 2*(q4*qd - q4d*q - hat_map(q).dot(qd))

    return Omega

def dcmtoaxisangle(R):
    angle = np.arccos((np.trace(R) - 1)/2)
    axis = 1/(2*np.sin(angle))*vee_map(R-R.T)
        
    return angle, axis

def axisangletodcm(angle, axis):
    ahat = hat_map(axis)
    R = np.eye(3,3) + np.sin(angle)*ahat + (1 - np.cos(angle))*ahat.dot(ahat)
    return R

def dcmdottoang_vel(R,Rdot):
    """Convert a rotation matrix to angular velocity
        
        w - angular velocity in inertial frame
        Omega - angular velocity in body frame
    """
    w = vee_map(Rdot.dot(R.T))
    Omega = vee_map(R.T.dot(Rdot))

    return (w, Omega)

def ang_veltodcmdot(R,Omega):
    """Convert angular velocity to DCM dot
        Omega - angular velocity defined in body frame 
    """
    Rdot = R.dot(hat_map(Omega))

    return Rdot

def ang_veltoaxisangledot(angle, axis, Omega):
    """Compute kinematics for axis angle representation

    """
    angle_dot = axis.dot(Omega)
    axis_dot = 1/2*(hat_map(axis) - 1/np.tan(angle/2) * hat_map(axis).dot(hat_map(axis))).dot(Omega)
    return angle_dot, axis_dot

def axisangledottoang_vel(angle,axis, angle_dot,axis_dot):
    """Convert axis angle represetnation to angular velocity in body frame

    """
    
    Omega = angle_dot*axis + np.sin(angle)*axis_dot - (1-np.cos(angle))*hat_map(axis).dot(axis_dot)
    return Omega

def test_rot_mat_in_special_orthogonal_group(R):
    np.testing.assert_array_almost_equal(np.linalg.det(R), 1, decimal=2)
    np.testing.assert_array_almost_equal(R.T.dot(R), np.eye(3,3), decimal=2)

def normalize(num_in, lower=0, upper=360, b=False):
    """Normalize number to range [lower, upper) or [lower, upper].

    Parameters
    ----------
    num : float
        The number to be normalized.
    lower : int
        Lower limit of range. Default is 0.
    upper : int
        Upper limit of range. Default is 360.
    b : bool
        Type of normalization. Default is False. See notes.

        When b=True, the range must be symmetric about 0.
        When b=False, the range must be symmetric about 0 or ``lower`` must
        be equal to 0.

    Returns
    -------
    n : float
        A number in the range [lower, upper) or [lower, upper].

    Raises
    ------
    ValueError
      If lower >= upper.

    Notes
    -----
    If the keyword `b == False`, then the normalization is done in the
    following way. Consider the numbers to be arranged in a circle,
    with the lower and upper ends sitting on top of each other. Moving
    past one limit, takes the number into the beginning of the other
    end. For example, if range is [0 - 360), then 361 becomes 1 and 360
    becomes 0. Negative numbers move from higher to lower numbers. So,
    -1 normalized to [0 - 360) becomes 359.

    When b=False range must be symmetric about 0 or lower=0.

    If the keyword `b == True`, then the given number is considered to
    "bounce" between the two limits. So, -91 normalized to [-90, 90],
    becomes -89, instead of 89. In this case the range is [lower,
    upper]. This code is based on the function `fmt_delta` of `TPM`.

    When b=True range must be symmetric about 0.

    Examples
    --------
    >>> normalize(-270,-180,180)
    90.0
    >>> import math
    >>> math.degrees(normalize(-2*math.pi,-math.pi,math.pi))
    0.0
    >>> normalize(-180, -180, 180)
    -180.0
    >>> normalize(180, -180, 180)
    -180.0
    >>> normalize(180, -180, 180, b=True)
    180.0
    >>> normalize(181,-180,180)
    -179.0
    >>> normalize(181, -180, 180, b=True)
    179.0
    >>> normalize(-180,0,360)
    180.0
    >>> normalize(36,0,24)
    12.0
    >>> normalize(368.5,-180,180)
    8.5
    >>> normalize(-100, -90, 90)
    80.0
    >>> normalize(-100, -90, 90, b=True)
    -80.0
    >>> normalize(100, -90, 90, b=True)
    80.0
    >>> normalize(181, -90, 90, b=True)
    -1.0
    >>> normalize(270, -90, 90, b=True)
    -90.0
    >>> normalize(271, -90, 90, b=True)
    -89.0
    """
    if lower >= upper:
        ValueError("lower must be lesser than upper")
    if not b:
        if not ((lower + upper == 0) or (lower == 0)):
            raise ValueError('When b=False lower=0 or range must be symmetric about 0.')
    else:
        if not (lower + upper == 0):
            raise ValueError('When b=True range must be symmetric about 0.')

    # abs(num + upper) and abs(num - lower) are needed, instead of
    # abs(num), since the lower and upper limits need not be 0. We need
    # to add half size of the range, so that the final result is lower +
    # <value> or upper - <value>, respectively.
    if not hasattr(num_in, "__iter__"):
        num_in = np.asarray([num_in], dtype=np.float)

    res  = []
    for num in num_in:
        if not b:
            if num > upper or num == lower:
                num = lower + abs(num + upper) % (abs(lower) + abs(upper))
            if num < lower or num == upper:
                num = upper - abs(num - lower) % (abs(lower) + abs(upper))

            res.append(lower if num == upper else num)
        else:
            total_length = abs(lower) + abs(upper)
            if num < -total_length:
                num += ceil(num / (-2 * total_length)) * 2 * total_length
            if num > total_length:
                num -= floor(num / (2 * total_length)) * 2 * total_length
            if num > upper:
                num = total_length - num
            if num < lower:
                num = -total_length - num

            res.append(num)

    return np.asarray(res, dtype=np.float)


def unit_vector(q):
    r"""Unit vector in direction of q

    qhat = unit_vector(q)

    Parameters
    ----------
    q : (n,) numpy array
        Vector in Rn

    Returns
    -------
    qhat : (n,) numpy array
        The unit vector in the direction of q

    Notes
    -----
    You may include some math:

    .. math:: \hat q  = \frac{q}{\norm{q}}

    Author
    ------
    Shankar Kulumani		GWU		skulumani@gwu.edu
    """
    return q / np.linalg.norm(q)

if __name__ == "__main__":
    pass



"""n-Sphere operations

This module holds some useful functions for dealing with elements of n-spheres.
Most usually we tend to deal with the 1-sphere and the 2-sphere.

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np

from kinematics import attitude

def rand(n, seed=9):
    """Random vector from the n-Sphere

    This function will return a random vector which is an element of the n-Sphere.
    The n-Sphere is defined as a vector in R^n with a norm of one. 

    Basically, we'll find a random vector in R^n and normalize it. 
    This is somewhat like mapping a cube to a sphere but hopefully it doesn't cause any
    issues in the future.

    Parameters
    ----------
    None

    Returns
    -------
    rvec 
        Random (n+1,) numpy vector with a norm of 1

    """
    rs = np.random.RandomState(seed)
    rvec = rs.rand(n+1)
    rvec = rvec / np.linalg.norm(rvec)
    return rvec

def tan_rand(q, seed=9):
    """Find a random vector in the tangent space of the n sphere

    This function will find a random orthogonal vector to q.

    Parameters
    ----------
    q
        (n+1,) array which is in the n-sphere

    Returns
    -------
    qd
        (n+1,) array which is orthogonal to n-sphere and also random

    """
    # probably need a check in case we get a parallel vector
    rs = np.random.RandomState(seed)
    rvec = rs.rand(q.shape[0])

    qd = np.cross(rvec, q)
    qd = qd / np.linalg.norm(qd)

    while np.dot(q, qd) > 1e-6:
        rvec = rs.rand(q.shape[0])
        qd = np.cross(rvec, q)
        qd = qd / np.linalg.norm(qd)

    return qd

def perturb_vec(q, cone_half_angle=2):
    r"""Perturb a vector randomly

    qp = perturb_vec(q, cone_half_angle=2)

    Parameters
    ----------
    q : (n,) numpy array
        Vector to perturb
    cone_half_angle : float
        Maximum angle to perturb the vector in degrees

    Returns
    -------
    perturbed : (n,) numpy array
        Perturbed numpy array

    Author
    ------
    Shankar Kulumani		GWU		skulumani@gwu.edu

    References
    ----------

    .. [1] https://stackoverflow.com/questions/2659257/perturb-vector-by-some-angle

    """
    rand_vec = tan_rand(q)
    cross_vector = attitude.unit_vector(np.cross(q, rand_vec))

    s = np.random.uniform(0, 1, 1)
    r = np.random.uniform(0, 1, 1)

    h = np.cos(np.deg2rad(cone_half_angle))

    phi = 2 * np.pi * s
    z = h + ( 1- h) * r
    sinT = np.sqrt(1 - z**2)
    x = np.cos(phi) * sinT
    y = np.sin(phi) * sinT

    perturbed = rand_vec * x + cross_vector * y + q * z
    
    return perturbed



"""This will test our functions on the n-sphere

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np

from kinematics import sphere

def test_two_sphere_random_vector_norm():
    vec = sphere.rand(2)
    np.testing.assert_almost_equal(np.linalg.norm(vec), 1)

def test_two_sphere_tangent_vector():
    vec = sphere.rand(2)
    vecd = sphere.tan_rand(vec, 9)

    np.testing.assert_almost_equal(np.dot(vec, vecd), 0)

class TestPerturbedVector():

    q = sphere.rand(2)
    half_angle = 5
    qp = sphere.perturb_vec(q, half_angle)

    def test_angle(self):
        """Ensure the angle is always less than half_angle
        """
        angle = np.arccos(np.dot(self.q, self.qp)) * 180 / np.pi
        np.testing.assert_array_less(angle, self.half_angle)

    def test_norm(self):
        np.testing.assert_allclose(np.linalg.norm(self.qp), 1)
    



"""
    Copyright (C) 2017  GWU Flight Dynamics and Control Laboratory 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
from kinematics import attitude


class TestHatAndVeeMap():
    x = np.random.rand(3)
    y = np.random.rand(3)

    def test_hat_map_zero(self):
        np.testing.assert_allclose(attitude.hat_map(self.x).dot(self.x), np.zeros(3))

    def test_hat_map_skew_symmetric(self):
        np.testing.assert_allclose(attitude.hat_map(self.x).T, -attitude.hat_map(self.x))

    def test_hat_vee_map_inverse(self):
        np.testing.assert_allclose(attitude.vee_map(attitude.hat_map(self.x)), self.x)

    def test_hat_map_cross_product(self):
        np.testing.assert_allclose(attitude.hat_map(self.x).dot(self.y), np.cross(self.x, self.y))
        np.testing.assert_allclose(attitude.hat_map(self.x).dot(self.y), -attitude.hat_map(self.y).dot(self.x))      


class TestEulerRot():
    angle = (2*np.pi - 0) * np.random.rand(1) + 0

    def test_rot1_orthogonal(self):
        mat = attitude.rot1(self.angle)
        np.testing.assert_array_almost_equal(mat.T.dot(mat), np.eye(3, 3))

    def test_rot2_orthogonal(self):
        mat = attitude.rot2(self.angle)
        np.testing.assert_array_almost_equal(mat.T.dot(mat), np.eye(3, 3))

    def test_rot3_orthogonal(self):
        mat = attitude.rot3(self.angle)
        np.testing.assert_array_almost_equal(mat.T.dot(mat), np.eye(3, 3))

    def test_rot1_determinant_1(self):
        mat = attitude.rot1(self.angle)
        np.testing.assert_allclose(np.linalg.det(mat), 1)

    def test_rot2_determinant_1(self):
        mat = attitude.rot2(self.angle)
        np.testing.assert_allclose(np.linalg.det(mat), 1)

    def test_rot3_determinant_1(self):
        mat = attitude.rot3(self.angle)
        np.testing.assert_allclose(np.linalg.det(mat), 1)

    def test_rot1_orthogonal_row(self):
        mat = attitude.rot1(self.angle, 'r')
        np.testing.assert_array_almost_equal(mat.T.dot(mat), np.eye(3, 3))

    def test_rot2_orthogonal_row(self):
        mat = attitude.rot2(self.angle, 'r')
        np.testing.assert_array_almost_equal(mat.T.dot(mat), np.eye(3, 3))

    def test_rot3_orthogonal_row(self):
        mat = attitude.rot3(self.angle, 'r')
        np.testing.assert_array_almost_equal(mat.T.dot(mat), np.eye(3, 3))

    def test_rot1_determinant_1_row(self):
        mat = attitude.rot1(self.angle, 'r')
        np.testing.assert_allclose(np.linalg.det(mat), 1)

    def test_rot2_determinant_1_row(self):
        mat = attitude.rot2(self.angle, 'r')
        np.testing.assert_allclose(np.linalg.det(mat), 1)

    def test_rot3_determinant_1_row(self):
        mat = attitude.rot3(self.angle, 'r')
        np.testing.assert_allclose(np.linalg.det(mat), 1)



class TestEulerRot90_column():
    angle = np.pi/2
    b1 = np.array([1, 0, 0])
    b2 = np.array([0, 1, 0])
    b3 = np.array([0, 0, 1])

    R1 = attitude.rot1(angle, 'c')
    R2 = attitude.rot2(angle, 'c')
    R3 = attitude.rot3(angle, 'c')
    
    R1_row = attitude.rot1(angle, 'r')
    R2_row = attitude.rot2(angle, 'r')
    R3_row = attitude.rot3(angle, 'r')
    
    def test_rot1_transpose(self):
        np.testing.assert_allclose(self.R1_row, self.R1.T)

    def test_rot2_transpose(self):
        np.testing.assert_allclose(self.R2_row, self.R2.T)

    def test_rot3_transpose(self):
        np.testing.assert_allclose(self.R3_row, self.R3.T)

    def test_rot1_90_b1(self):
        np.testing.assert_array_almost_equal(self.R1.dot(self.b1), self.b1)

    def test_rot1_90_b2(self):
        np.testing.assert_array_almost_equal(self.R1.dot(self.b2), self.b3)

    def test_rot1_90_b3(self):
        np.testing.assert_array_almost_equal(self.R1.dot(self.b3), -self.b2)

    def test_rot2_90_b1(self):
        np.testing.assert_array_almost_equal(self.R2.dot(self.b1), -self.b3)

    def test_rot2_90_b2(self):
        np.testing.assert_array_almost_equal(self.R2.dot(self.b2), self.b2)

    def test_rot2_90_b3(self):
        np.testing.assert_array_almost_equal(self.R2.dot(self.b3), self.b1)

    def test_rot3_90_b1(self):
        np.testing.assert_array_almost_equal(self.R3.dot(self.b1), self.b2)

    def test_rot3_90_b2(self):
        np.testing.assert_array_almost_equal(self.R3.dot(self.b2), -self.b1)

    def test_rot3_90_b3(self):
        np.testing.assert_array_almost_equal(self.R3.dot(self.b3), self.b3)
    
class TestEulerRotInvalid():
    R1 = attitude.rot1(0, 'x')
    R2 = attitude.rot2(0, 'x')
    R3 = attitude.rot3(0, 'x')
    invalid = 1
    def test_rot1_invalid(self):
        np.testing.assert_allclose(self.R1, self.invalid)

    def test_rot2_invalid(self):
        np.testing.assert_allclose(self.R2, self.invalid)

    def test_rot3_invalid(self):
        np.testing.assert_allclose(self.R3, self.invalid)

class TestExpMap():
    angle = (np.pi - 0) * np.random.rand(1) + 0
    axis = np.array([1, 0, 0])
    R = attitude.rot1(angle)

    def test_axisangletodcm(self):
        np.testing.assert_array_almost_equal(attitude.rot1(self.angle), attitude.axisangletodcm(self.angle, self.axis))

    def test_dcmtoaxisangle(self):
        angle, axis = attitude.dcmtoaxisangle(self.R)
        np.testing.assert_array_almost_equal(angle, self.angle)
        np.testing.assert_array_almost_equal(axis, self.axis)

class TestQuaternion():
    angle = (2*np.pi - 0) * np.random.rand(1) + 0
    dcm = attitude.rot1(angle)
    quat = attitude.dcmtoquat(dcm)

    dcm_identity = np.eye(3,3)
    quat_identity = attitude.dcmtoquat(dcm_identity)

    def test_dcmtoquaternion_unit_norm(self):
        np.testing.assert_almost_equal(np.linalg.norm(self.quat), 1)

    def test_quaternion_back_to_rotation_matrix(self):
        np.testing.assert_array_almost_equal(attitude.quattodcm(self.quat),
                                            self.dcm)

    def test_identity_quaternion_scalar_one(self):
        np.testing.assert_equal(self.quat_identity[-1], 1)
    
    def test_identity_quaternion_vector_zero(self):
        np.testing.assert_array_almost_equal(self.quat_identity[0:3], np.zeros(3))

class TestNormalize():
    
    def test_lower_limit_circle(self):
        expected = -180 
        actual = attitude.normalize(-180, -180, 180)
        np.testing.assert_almost_equal(actual, expected)

    def test_upper_limit_circle(self):
        expected = -180
        actual = attitude.normalize(180, -180, 180)
        np.testing.assert_almost_equal(actual, expected)

    def test_past_upper_limit_circle(self):
        expected = -179
        actual = attitude.normalize(181, -180, 180)
        np.testing.assert_almost_equal(actual, expected)

    def test_past_lower_limit_circle(self):
        expected = 179
        actual = attitude.normalize(-181, -180, 180)
        np.testing.assert_almost_equal(actual, expected)
    
    def test_vectorized_interior(self):
        angles = np.linspace(10, 50, 10)
        expected = angles
        actual = attitude.normalize(angles, 0, 180)
        np.testing.assert_allclose(actual, expected)

class TestUnitVector():
    
    q = np.array([5, 0, 0])
    qhat_true = np.array([1, 0, 0])
    qhat = attitude.unit_vector(q)

    def test_norm(self):
        np.testing.assert_allclose(np.linalg.norm(self.qhat), 1)

    def test_unit_vector(self):
        np.testing.assert_allclose(self.qhat, self.qhat_true)
